import re
from collections import defaultdict
from pathlib import Path
from pybtex import database
from pybtex.database.input import BaseParser
from pybtex.database import BibliographyData, Entry, Person
from pybtex.utils import OrderedCaseInsensitiveDict
import csv

data_dir = Path(__file__).parent/"data"

class RISParser(BaseParser):
    """
    Parser for RIS files.
    """
    default_suffix = '.ris'
    unicode_io = True

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.ris_type_to_bibtex = {}
        self.ris_type_description = {}

        with open(data_dir/"types.csv") as f:
            reader = csv.reader(f, delimiter=',')
            
            #skip the header
            next(reader, None)
            for row in reader:
                ris_types = row[0]
                description = row[1]
                bibtex_type = row[2]

                for ris_type in ris_types.split(", "):
                    self.ris_type_to_bibtex[ris_type] = bibtex_type
                    self.ris_type_description[ris_type] = description

    def parse_stream(self, stream):
        text = stream.read()
        return self.parse_string(text)

    def process_entry(self, entry_text):
        """
        https://github.com/aurimasv/translators/wiki/RIS-Tag-Map
        """
        # Read entry into dictionary
        ris_dict = defaultdict(list)
        for line in entry_text.split('\n'):
            m = re.match(r"^([A-Z0-9]{2})\s*-\s*(.*)$", line.strip())
            if not m:
                continue

            code = m.group(1)
            value = m.group(2)
            
            ris_dict[code].append(value)
        
        # Read type of entry
        ris_type = ris_dict.pop("TY", ["GEN"])
        ris_type = ris_type[0]
        ris_description = None
        if ris_type in self.ris_type_description.values():
            index = list(self.ris_type_description.values()).index(ris_type)
            ris_description = ris_type
            ris_type = list(self.ris_type_description.keys())[index]
            
        ris_type = ris_type.upper()
        if not ris_description:
            ris_description = self.ris_type_description[ris_type]

        # Create Entry object
        bibtex_type = self.ris_type_to_bibtex.get(ris_type, "misc")
        entry = Entry(bibtex_type)
        entry.fields["type"] = ris_description
        
        # Read People
        def add_person(code, role):
            names = ris_dict.pop(code, [])
            for name in names:
                person = Person(name)
                entry.add_person(person, role)

        add_person("AU", "author")
        add_person("A1", "author")
        editor_role = ["ANCIENT", "BLOG", "BOOK", "CHAP", "CLSWK", "COMP", "DATA", "CPAPER", "CONF", "DICT", "EDBOOK", "EBOOK", "ECHAP", "ENCYC", "MAP", "MUSIC", "MULTI", "RPRT", "SER", "UNPB", "ELEC"]
        add_person("A2", "editor" if ris_type in editor_role else "author")
        editor_role += ["ADVS", "SLIDE", "SOUND", "VIDEO"]
        add_person("A3", "editor" if ris_type in editor_role else "author")
        add_person("A4", "editor" if ris_type in editor_role else "author")        
        add_person("ED", "editor")        

        # Read Other Fields
        def add_field(code, bibtex_field, delimiter=", "):
            values = ris_dict.pop(code, [])
            for value in values:
                if bibtex_field in entry.fields:
                    entry.fields[bibtex_field] += f"{delimiter}{value}"
                else:
                    entry.fields[bibtex_field] = value

        add_field("TI", "title")
        if "title" not in entry.fields:
            add_field("T1", "title")
        add_field("JO", "journal")

        bibtex_t2 = None
        if ris_type in ["ABST","INPR","JFULL","JOUR", "EJOUR"]:
            bibtex_t2 = "journal"
        elif ris_type in ["ANCIENT","CHAP", "ECHAP"]:
            bibtex_t2 = "booktitle"
        elif ris_type in ["BOOK","CTLG", "CLSWK", "COMP", "DATA", "MPCT", "MAP", "MULTI", "RPRT", "UNPB", "ELEC"]:
            bibtex_t2 = "series"
        if bibtex_t2:
            add_field("T2", bibtex_t2)
            
        bibtex_bt = bibtex_t2
        if ris_type == "BOOK":
            bibtex_bt = "title"
        if bibtex_bt:
            add_field("BT", bibtex_bt)

        bibtex_t3 = None
        if ris_type in ["BOOK","CTLG", "CLSWK", "COMP", "DATA", "MPCT", "MAP", "MULTI", "RPRT", "UNPB", "ELEC", "ADVS", "SLIDE", "SOUND", "VIDEO", "CHAP", "CONF", "DATA", "EBOOK", "ECHAP", "GOVDOC", "MUSIC", "SER"]:
            bibtex_t3 = "series"
        if bibtex_t3:
            add_field("T3", bibtex_t3)
            
        add_field("PB", "publisher")
        add_field("CY", "address")
        add_field("IS", "number")
        add_field("DO", "doi")
        add_field("VL", "volume")
        add_field("SP", "pages")
        add_field("UR", "url")
        add_field("KW", "keywords", delimiter=" | ")
        add_field("N1", "note", delimiter=" | ")
        add_field("N2", "note", delimiter=" | ")
        add_field("PY", "year")
        add_field("AB", "abstract")
        if "year" not in entry.fields:
            add_field("Y1", "year")

        # Read ISBN or ISSN
        serial_numbers = ris_dict.pop("SN", [])
        for serial_number in serial_numbers:
            sn_digits = re.sub(r"\D","", serial_number)
            sn_field = "issn" if len(sn_digits) == 8 else "isbn"
            entry.fields[sn_field] = serial_number

        entry_key = ris_dict.pop("ID", [None])
        entry_key = entry_key[0]

        # Check if DA field could be a month
        dates = ris_dict.get("DA", [])
        for date in dates:
            digital_month = date.isdigit() and (1 <= int(date) <= 12)
            text_month = date[:3].lower() in ["jan", "feb", "mar", "apr", "may", "jun", "jul", "aug", "sep", "oct", "nov", "dec"]
            if digital_month or text_month:
                entry.fields["month"] = date
                ris_dict.pop("DA")
                
        end_pages = ris_dict.pop("EP", [])
        for end_page in end_pages:
            if "pages" in entry.fields:
                entry.fields["pages"] += f"--{end_page}"
            else:
                entry.fields["pages"] = end_page

        # Add the remaining fields with the RIS code as the field name
        ris_dict.pop("ER", None)
        for code, values in ris_dict.items():
            entry.fields[code] = ", ".join(values)

        
        if not entry_key:
            people = [x[0] for x in entry.persons.values()]
            if people:
                first_author = people[0]
                entry_key = "-".join(first_author.last_names).replace(" ", ".")
            elif "title" in entry.fields:
                TRUNCATE_TITLE = 20
                entry_key = entry.fields["title"].replace(" ", ".")[:TRUNCATE_TITLE]
            else:
                entry_key = "Unknown"

            if "year" in entry.fields:
                entry_key += entry.fields["year"]

        return entry_key,entry

    def parse_string(self, text):
        self.unnamed_entry_counter = 1
        self.command_start = 0

        entry_texts = re.split(r"ER\s+-", text)
        entries = (self.process_entry(t) for t in entry_texts if t.strip())
        self.data.add_entries(entries)
        return self.data


class SuffixParser(BaseParser):
    """
    A parser which chooses the parser based on the suffix of each file given to it.
    """
    def parse_file(self, filename, file_suffix=None):
        if file_suffix is not None:
            filename = str(filename) + file_suffix

        file_data = database.parse_file(filename)
        self.data.add_entries(file_data.entries.items())
        if file_data._preamble:
            self.data._preamble.extend(file_data._preamble)

        return file_data




# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pybtexris']

package_data = \
{'': ['*'], 'pybtexris': ['data/*']}

install_requires = \
['pybtex>=0.24.0,<0.25.0']

entry_points = \
{'pybtex.database.input': ['ris = pybtexris:RISParser',
                           'suffix = pybtexris:SuffixParser'],
 'pybtex.database.input.suffixes': ['.ris = pybtexris:RISParser']}

setup_kwargs = {
    'name': 'pybtexris',
    'version': '0.1.1',
    'description': 'A pybtex plugin for working with RIS files.',
    'long_description': '============\npybtexris\n============\n\n.. start-badges\n\n|pipline badge| |coverage badge| |black badge| |git3moji badge|\n\n.. |pipline badge| image:: https://github.com/rbturnbull/pybtexris/actions/workflows/coverage.yml/badge.svg\n    :target: https://github.com/rbturnbull/pybtexris/actions\n    \n.. |black badge| image:: https://img.shields.io/badge/code%20style-black-000000.svg\n    :target: https://github.com/psf/black\n    \n.. |coverage badge| image:: https://img.shields.io/endpoint?url=https://gist.githubusercontent.com/rbturnbull/665c8745fce7077155f99ad694a7e762/raw/coverage-badge.json\n    :target: https://rbturnbull.github.io/pybtexris/\n\n.. |git3moji badge| image:: https://img.shields.io/badge/git3moji-%E2%9A%A1%EF%B8%8F%F0%9F%90%9B%F0%9F%93%BA%F0%9F%91%AE%F0%9F%94%A4-fffad8.svg\n    :target: https://robinpokorny.github.io/git3moji/\n\n.. end-badges\n\nA pybtex plugin for working with RIS files.\n\nInstallation\n============\n\nInstall pybtexris from PyPI using pip::\n\n    pip install pybtexris\n\nCommand-line usage\n==================\n\nTo convert an RIS file to another format, use the ``pybtex-convert`` command. For example::\n\n    pybtex-convert bibliography.ris bibliography.bib\n\nThe extension of the output file must be supported by ``pybtex`` or an associated plugin.\n\nTo format an RIS file into a human-readable bibliography, use the pybtex-format command. For example::\n\n    pybtex-format bibliography.ris bibliography.txt\n\nFor more information, see `the documentation for pybtex <https://docs.pybtex.org/cmdline.html>`_.\n\nProgrammatic usage\n==================\n\nRIS files can be formatted into a human-readable bibliography as a string as follows:\n\n.. code-block:: python\n\n    from pybtex import format_from_file\n    bibliography_string = format_from_file(\n        "path/to/file.ris", \n        style="plain", \n        output_backend="plaintext",\n        bib_format="ris",\n    )\n\nMultiple RIS files can be formatted in a similar way:\n\n.. code-block:: python\n\n    from pybtex import format_from_files\n    bibliography_string = format_from_files(\n        ["path/to/file1.ris", "path/to/file2.ris"],\n        style="plain", \n        output_backend="plaintext",\n        bib_format="ris",\n    )\n\nSo that RIS files can be combined with bibliography files of other formats (such as BibTeX), \n`pybtexris` also adds `SuffixParser` to the list of plugins which pybtex can use.\nThe user just needs to give ``suffix`` as the argument to ``bib_format``.\n\n.. code-block:: python\n\n    from pybtex import format_from_files\n    result = format_from_files(\n        ["path/to/file1.ris", "path/to/file2.bib"],\n        style="plain", \n        output_backend="plaintext",\n        bib_format="suffix",\n    )\n\nThe parsers for the files for other formats need to be registered on the ``pybtex.database.input.suffixes``\nentry point as discussed pybtex `plugin documentation <https://docs.pybtex.org/api/plugins.html>`_.\nTo combine with NBIB citation files, please use the `pybtexnbib <https://github.com/rbturnbull/pybtexnbib>`_ plugin.\n\nFor more information on programmatic use of pybtex, \nsee `the documentation of the Python API of pybtex <https://docs.pybtex.org/api/index.html>`_.\n\nCredit\n==================\n\nRobert Turnbull (Melbourne Data Analytics Platform, University of Melbourne)',
    'author': 'Robert Turnbull',
    'author_email': 'robert.turnbull@unimelb.edu.au',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/rbturnbull/pybtexris',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7.1,<4.0.0',
}


setup(**setup_kwargs)
